/*
** Module   : P2JH2Dialect.java
** Abstract : Implementation of P2JDialect for a H2 backend
**
** Copyright (c) 2008-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20080223   @37519 Created initial version. Adds P2JDialect
**                           method implementations to extend Hibernate's
**                           H2Dialect.
** 002 ECF 20080611   @38701 Added isCaseInsensitiveColumn(). Provides a
**                           rough test for case-insensitivity, based only
**                           on an expression which contains a column
**                           name, such as might be extracted from index
**                           metadata. Commented out experimental code.
** 003 SVL 20080623   @38890 Column type for the VARBINARY type was
**                           changed to "binary" (instead of
**                           "binary($l)").
** 004 ECF 20080805   @39315 Added getCreateSequenceString(). Generates
**                           DDL to create a sequence starting at a
**                           particular value.
** 005 CA  20080815   @39443 Added isConnectionError method - it checks if
**                           a given exception represents a connection
**                           error.
** 006 ECF 20080930   @40002 Added isQueryRangeParameterInlined().
**                           Indicates whether query substitution
**                           parameters within range checks should be
**                           inlined.
** 007 ECF 20081015   @40097 Added formatDate(). Formats a date into a
**                           string suitable for use in an SQL statement.
** 008 ECF 20090310   @41500 Added inlineLimit(). Enables inlining the
**                           limit/max for records returned by a query
**                           into the SELECT statement. Currently it is
**                           unimplemented.
** 009 SVL 20090810   @43574 Added requiresExplicitCastInsideTernary().
**                           Determines whether substitution parameters should
**                           be explicitly casted to their datatypes inside
**                           "then ? else ?" statement.
** 010 GES 20100701          Added a type name override to allow honoring of
**                           precision and scale in numeric types.
** 011 CA  20101026          Added methods to generate the DROP/CREATE INDEX
**                           statements in a dialect specific form. Also, 
**                           added method to generate the ALTER TABLE statement
**                           which sets the formula for the computed columns.
**                           Added methods to prepare/process string metadata 
**                           query parameters and results, to get the database
**                           prepare DDL statements and collation statement.
** 012 CA  20101202          Added method to build the URL for a remote 
**                           connection. Removed the "h2.sortNullsHigh" 
**                           setting, as it caused regressions; more testing 
**                           in a 4GL environment is needed, to determine 
**                           exactly how 4GL sorts the unknown value with the
**                           temporary and permanent databases.
** 013 SVL 20110622          Added usesSingleBackslash.
** 014 OM  20121120          Added DDL/SQL generation support for sequences.
** 015 SVL 20130331          Upgraded to Hibernate 4.
** 016 ECF 20130510          Fixed sequence-related methods to eliminate use of case-sensitive
**                           sequence names.
** 017 OM  20130712          Added if exists clause to drop sequence statement.
*/

package com.goldencode.p2j.persist.dialect;

import java.net.InetAddress;
import java.net.URI;
import java.sql.*;
import java.sql.Date;
import java.text.NumberFormat;
import java.util.*;
import java.util.List;

import org.hibernate.*;
import org.hibernate.cfg.Configuration;
import org.hibernate.dialect.*;
import org.hibernate.engine.spi.*;
import org.hibernate.id.factory.*;
import org.hibernate.mapping.*;
import org.hibernate.type.Type;

import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.embed.H2Helper;
import com.goldencode.p2j.util.*;

/**
 * Extends Hibernate's <code>H2Dialect</code> class to add method
 * implementations required by the {@link P2JDialect} interface.
 */
public class P2JH2Dialect
extends H2Dialect
implements P2JDialect
{
//   /** Logger */
//   private static final Log LOG = LogFactory.getLog(P2JH2Dialect.class);
   
   /** Prefix prepended to computed column/property names */
   private static final String CC_PREFIX = "__";
   
//   /** Actual minimum interval in milliseconds between analyze commands */
//   private static final long analyzeInterval = 60000;
   
//   /** Debug level logging enabled? */
//   private static final boolean debug = LOG.isDebugEnabled();
   
   /**
    * Default constructor.  This constructor is for Hibernate's use.  Use code
    * should not construct instances of this class.
    * <p>
    * We provide this constructor to override the superclass' mapping of 
    * <code>Types.VARCHAR</code>, in order to force this type to map to
    * PostgreSQL's <code>text</code> data type.
    */
   public P2JH2Dialect()
   {
      super();
      
      // Override superclass' VARCHAR and VARBINARY mappings, since we always
      // use the unbounded variants of these data types.
      registerColumnType(Types.VARCHAR, "varchar");
      registerColumnType(Types.VARBINARY, "binary");
      
      // Override superclass' NUMERIC mapping, since the Hibernate H2 dialect
      // doesn't honor precision and scale.
      registerColumnType(Types.NUMERIC, "numeric($p,$s)");
   }
   
   /**
    * Indicate whether this dialect requires computed columns to participate
    * within index definitions, instead of embedded expressions.
    * 
    * @return  <code>true</code> to indicate computed columns are required.
    */
   public boolean needsComputedColumns()
   {
      return true;
   }
   
   /**
    * Get the prefix, if any, which should be prepended to computed column
    * and/or property names.
    * 
    * @return  Computed column/property prefix.
    * 
    * @see     #needsComputedColumns()
    */
   public String getComputedColumnPrefix()
   {
      return CC_PREFIX;
   }
   
   /**
    * Indicate whether this dialect natively supports the overloading of user
    * defined functions by parameter number and data type.
    * 
    * @return  <code>false</code>.
    */
   public boolean supportsFunctionOverloading()
   {
      return false;
   }
   
   /**
    * Create a data object, the purpose of which is specific to a particular
    * dialect implementation, to be stored as a context-local object.  This
    * allows a dialect to establish context-local data which is stored by the
    * persistence runtime framework, without the overhead of a separate,
    * context-local variable.  This object will be handed back to the dialect
    * in callback methods when the services of certain hooks are required.
    * <p>
    * Since the context-local data is specific to a particular dialect
    * implementation, it is optional.  Dialect implementations which do not
    * have a need for this data should simply return <code>null</code>.
    * 
    * @param   persistence
    *          Provider of persistence services.
    * 
    * @return  Dialect-specific, context-local data, or <code>null</code>.
    */
   public Object createContext(Persistence persistence)
   {
      /*
      Database database = persistence.getDatabase();
      
      return (analyzeInterval > 0 &&
              DatabaseManager.TEMP_TABLE_DB.equals(database)
              ? (new Context())
              : null);
      */
      
      return null;
   }
   
   /**
    * A dialect-specific callback hook which is invoked by the persistence
    * runtime framework after a transaction is complete.  The Hibernate
    * session will have been closed by the time this method is invoked.
    * 
    * @param   persistence
    *          Provider of persistence services.
    * @param   context
    *          Optional, context-local data created previously by {@link
    *          #createContext}.
    * @param   rollback
    *          <code>true</code> if transaction ended with a rollback;
    *          <code>false</code> if transaction ended with a commit.
    * 
    * @throws  PersistenceException
    *          if any persistence-related error occurs.
    */
   public void postTransaction(Persistence persistence,
                               Object context,
                               boolean rollback)
   throws PersistenceException
   {
      /*
      if (context != null)
      {
         Context local = (Context) context;
         long now = System.currentTimeMillis();
         long diff = now - local.lastAnalyze;
         if (diff >= analyzeInterval)
         {
            int count = 0;
            try
            {
               Iterator<String> iter = TemporaryBuffer.activeTables();
               while (iter.hasNext())
               {
                  iter.next();
                  count++;
               }
               if (count > 0)
               {
                  persistence.executeSQL("analyze");
               }
            }
            finally
            {
               long end = System.currentTimeMillis();
               long elapsed =  end - now;
               local.analyzeTime += elapsed;
               if (debug && count > 0)
               {
                  StringBuilder buf = new StringBuilder("Analyze ");
                  buf.append(count);
                  buf.append(" temp table(s) [");
                  buf.append(diff);
                  buf.append(":");
                  buf.append(elapsed);
                  buf.append(":");
                  buf.append(local.analyzeTime);
                  buf.append("]");
                  LOG.debug(persistence.message(buf.toString()));
               }
               local.lastAnalyze = end;
            }
         }
      }
      */
   }
   
   /**
    * Retrieve the earliest possible date value which is valid for the
    * backing database.
    *
    * @return  <code>null</code>, indicating no particular lower limit on
    *          dates.
    */
   public Date getLowDate()
   {
      return null;
   }
   
   /**
    * Retrieve the latest possible date value which is valid for the backing
    * database.
    *
    * @return  <code>null</code>, indicating no particular upper limit on
    *          dates.
    */
   public Date getHighDate()
   {
      return null;
   }
   
   /**
    * Format a date object as a string which is appropriate for use in an SQL
    * statement.
    * 
    * @param   d
    *          Date to be formatted as a string.
    * 
    * @return  Formatted representation of the given date.
    */
   public String formatDate(date d)
   {
      NumberFormat nf = NumberFormat.getInstance(); 
      
      nf.setMinimumIntegerDigits(4);
      nf.setGroupingUsed(false);
      
      StringBuffer buf = new StringBuffer();
      int year = d.getYear();
      
      buf.append(nf.format(Math.abs(year)));
      buf.append('-');
      
      nf.setMinimumIntegerDigits(2);
      buf.append(nf.format(d.getMonth()));
      buf.append('-');
      buf.append(nf.format(d.getDay()));
      
      if (year < 0)
      {
         buf.append(" BC");
      }
      
      return buf.toString();
   }
   
   /**
    * Given a native column name as queried from database metadata, extract
    * the root name of a column.  For example, a column name as found within
    * an index in metadata might be represented as:
    * <pre>
    *    __my_column
    * <pre>
    * We are interested only in the <code>my_column</code> root name, so we
    * strip away the leading <code>__</code> prefix.
    *
    * @param   expr
    *          Column name as natively represented by the database dialect.
    *
    * @return  Root column name.
    */
   public String extractColumnName(String expr)
   {
      if (expr.startsWith(CC_PREFIX))
      {
         return expr.substring(CC_PREFIX.length());
      }
      
      return expr;
   }
   
   /**
    * Adjust a character column name appropriate to this dialect, such that it
    * can be inserted into an index.  This involves prepending a special
    * prefix to designate a computed column.
    * 
    * @param   name
    *          Base column name.
    * @param   ignoreCase
    *          <code>true</code> to cause case to be ignored, else
    *          <code>false</code>.
    * 
    * @return  Processed column name.
    */
   public String getProcessedCharacterColumnName(String name,
                                                 boolean ignoreCase)
   {
      StringBuilder buf = new StringBuilder(CC_PREFIX);
      buf.append(name);
      
      return buf.toString();
   }
   
   /**
    * Report whether the given column name represents an alias or an
    * expression which indicates the column is case insensitive.  Assume the
    * column represented by <code>name</code> is aleady determined to be the
    * appropriate data type for textual data.
    * 
    * @param   name
    *          Column name, alias, or expression.
    * 
    * @return  <code>true</code> if <code>name</code> represents
    *          case-insensitive data.
    */
   public boolean isCaseInsensitiveColumn(String name)
   {
      return name.startsWith(CC_PREFIX);
   }
   
   /**
    * Generate a DDL string which will create a global sequence with the given
    * starting value.
    * 
    * @param   name
    *          Sequence name.
    * @param   start
    *          Sequence starting value.
    * 
    * @return  Dialect-specific string to create the sequence, or
    *          <code>null</code> if not supported.
    */
   public String getCreateSequenceString(String name, long start)
   {
      StringBuilder buf = new StringBuilder(getCreateSequenceString(name));
      buf.append(" start with ");
      buf.append(start);
      
      return buf.toString();
   }

   /**
    * Generate a DDL string which will create a sequence with given properties.
    * 
    * @param   name
    *          Sequence name. Mandatory not null.
    * @param   init
    *          Sequence starting value. Mandatory(not null).
    * @param   increment
    *          Sequence incrementing value. Mandatory(not null).
    * @param   minVal
    *          Sequence minimum value. Optional (may be null).
    * @param   maxVal
    *          Sequence maximum value. Optional (may be null).
    * @param   cycle
    *          Sequence cycling property. Optional, if missing assumed not cycling.
    * 
    * @return  Dialect-specific string to create the sequence, or <code>null</code> if 
    *          not supported.
    */
   @Override
   public String getCreateSequenceString(String name,
                                         Long init,
                                         Long increment,
                                         Long minVal,
                                         Long maxVal,
                                         Boolean cycle)
   {
      // H2 create sequence syntax:
      // CREATE SEQUENCE [ IF NOT EXISTS ] newSequenceName
      //       [ START WITH long ]
      //       [ INCREMENT BY long ]
      //       [ CACHE long ]
      // H2 (at least in version 1.3.169) does not support min/max values
      //    nor cycle sequences, those features will be emulated by p2j

      StringBuilder buf = new StringBuilder(getCreateSequenceString(name, init));
      buf.append(" increment by ").append(increment);
      buf.append(" cache 1");
      
      return buf.toString();
   }

   /**
    * Generate a DDL string which will drop a sequence with given properties.
    * 
    * @param   name
    *          The name of the sequence to be dropped.
    *
    * @return  Dialect-specific string to drop the sequence, or
    *          <code>null</code> if not supported.
    */
   @Override
   public String getDropSequenceString(String name)
   {
      StringBuilder buf = new StringBuilder("drop sequence if exists ");
      buf.append(name);

      return buf.toString();
   }
   
   /**
    * Generate the appropriate select statement to reset the value of a 
    * sequence.
    * 
    * @param   sequenceName
    *          The name of the sequence.
    * @param   newVal
    *          The new value to be set.
    *
    * @return  Dialect-specific string to query the sequence, or
    *          <code>null</code> if not supported.
    */
   public String getSequenceSetValString(String sequenceName, long newVal)
   {
      StringBuilder sb = new StringBuilder();
      sb.append("alter sequence ");
      sb.append(sequenceName).append(" restart with ");
      sb.append(newVal);
      
      return sb.toString();
   }

   /**
    * Generate the appropriate select statement to retrieve the current 
    * value of a sequence.
    * 
    * @param   sequenceName
    *          The name of the sequence to be queried.
    *
    * @return  Dialect-specific string to query the sequence, or
    *          <code>null</code> if not supported.
    */
   @Override
   public String getSequenceCurrValString(String sequenceName)
   {
      // this feature is not supported by H2 directly, use a workaround
      StringBuilder buf = new StringBuilder(
         "select current_value from information_schema.sequences where upper(sequence_name) = '");
      buf.append(sequenceName.toUpperCase());
      buf.append("'");
      
      return buf.toString();
   }

   /**
    * Returns <code>true</code> if this dialect fully support the sequences as
    * 4GL language (most important: bounded values, cycle). If not fully
    * supported, additional code is needed in P2J sequence class
    * implementation to fix this.
    * 
    * @return false as H2 dialect does not fully support 4GL sequences
    */
   @Override
   public boolean hasFullSequenceSupport()
   {
      return false;
   }

   /**
    * Given an exception, this method will check if it indicates a database
    * connectivity error.
    * <p>
    * Currently, as the H2 database is used only for in-memory temp-tables, 
    * this method returns false.
    * 
    * @param   exc
    *          The exception to be checked.
    * 
    * @return  <code>true</code> if the exception indicates a database 
    *          connectivity error.
    */
   public boolean isConnectionError(Exception exc)
   {
      // TODO:  implement.
      return false;
   }

   /**
    * Indicate whether a query parameter which is part of a range check should
    * be inlined into the where clause, or whether it should be substituted
    * into a prepared statement.  This will result in an "inlined"
    * substitution parameter within a where clause.  For example, given a
    * where clause of:
    * <pre>
    *    where value >= ? and id = ?
    * </pre>
    * along with respective substitution parameters of 45 and 100, the
    * inlined version would read:
    * <pre>
    *    where value >= 45 and id = ?
    * </pre>
    * Only the <code>id</code> substitution parameter of 100 would be
    * substituted into the prepared statement.
    * <p>
    * Inlining in such a way can give some database dialects the opportunity
    * to choose a better query plan, than if the parameter being tested with
    * a range check were to be substituted at statement execution time.
    * <p>
    * Note:  this serves as a hint which may be overridden by runtime logic or
    * other configuration settings.
    * 
    * @return  <code>true</code> to inline range check parameters;
    *          <code>false</code> to use normal parameter substitution.
    */
   public boolean isQueryRangeParameterInlined()
   {
      return false;
   }

   /**
    * Enable/disable limit clause inlining.  Inlining the limit clause will
    * cause the limit/max value to be hard-coded into the SQL statement,
    * rather than being bound as a substitution parameter.
    * <p>
    * Dialects are free to treat this method as a hint.  This implementation
    * does nothing.
    * 
    * @param   on
    *          <code>true</code> to enable inlining;  <code>false</code> to
    *          disable it.
    */
   public void inlineLimit(boolean on)
   {
   }

   /**
    * Determines whether substitution parameters should be explicitly casted
    * to their datatypes inside "then ? else ?" statement.
    *
    * @return <code>true</code>.
    */
   public boolean requiresExplicitCastInsideTernary()
   {
      return true;
   }

   /**
    * Determines whether dialect uses single backslash (e.g. '12\34') into
    * string literals rather than double backslash (e.g. '12\\34').
    *
    * @return <code>true</code>
    */
   public boolean usesSingleBackslash()
   {
      return true;
   }

//   /**
//    * Context local data needed by this object to perform dialect-specific
//    * processing.  In this case, we track the last time a temp table analyze
//    * was performed and how much accumulated time we have spent performing
//    * such analyze commands in this context.
//    */
//   private static class Context
//   {
//      /** Time of last temp table analyze */
//      private long lastAnalyze = System.currentTimeMillis();
//      
//      /** Accumulated time spent analyzing temp tables */
//      private long analyzeTime = 0L;
//   }

   /**
    * Build the DDL to drop an index based upon the specified definition.
    * The SQL is generated by using the database-specific dialect.
    *
    * @param   index
    *          Definition of the index to be dropped.
    *
    * @return  The DDL to drop the given index.
   */
   public String getDropIndexString(P2JIndex index)
   {
      return "drop index " + index.getName();
   }

   /**
    * Build the DDL to create an index based upon the specified definition.
    * The SQL is generated by using a database-specific dialect.
    * <p>
    * When Hibernate decides to support "asc/desc" clauses, this method needs 
    * to be modified, as it appends the " desc" suffix to the column name,
    * if the column sort is decending. 
    * 
    * @param   index
    *          Definition of the index to be created.
    * @param   unique
    *          If <code>true</code>, create the index as a unique index,
    *          <code>false</code> to omit the unique constraint.
    * @param   toFile
    *          <code>true</code> if the string will be output to a file or
    *          <code>false</code> if it will be directly submitted via JDBC.
    *
    * @return  The DDL which was created.
   */
   public String getCreateIndexString(P2JIndex index, 
                                      boolean  unique, 
                                      boolean  toFile)
   {
      // Add primary key column to any index which is not explicitly or
      // implicitly unique.
      // TODO:  move this to import ruleset so we don't hardcode primary
      // key column name.
      if (!index.isUnique())
      {
         index.addComponent("id");
      }
      
      // Extract index information into a Hibernate-friendly form.
      String tabName = index.getTable();
      Table table = new Table(tabName);
      List<Column> columns = new ArrayList<Column>();
      Iterator<P2JIndexComponent> iter = index.components();
      while (iter.hasNext())
      {
         P2JIndexComponent next = iter.next();
         String colName = next.getName();
         if (next.isCharType())
         {
            colName = getComputedColumnPrefix() + colName;
         }
         
         // to allow asc/desc clauses for the index, we manipulate the column
         // name to add the "desc" direction, if needed; this is safe as the
         // H2 dialect doesn't have quoted columns, so the output DDL will be
         // without quotes
         if (next.isDescending())
         {
            colName = colName + " desc";
         }
         
         columns.add(new Column(colName));
      }

      // Compose the string.
      String createString =
         Index.buildSqlCreateIndexString(this,
                                         index.getName(),
                                         table,
                                         columns.iterator(),
                                         unique,
                                         null,
                                         null);
      
      return createString;
   }
   
   /**
    * Build the DDL to set the formula for the computed columns associated 
    * with an indexed, character column. 
    * <p>
    * The syntax to set the formula for the computed column is:
    * <code>
    * alter table [table] alter column [column] [type] as [formula]
    * </code>
    * 
    * @param   config
    *          The hibernate's configuration object.
    * @param   implClassName
    *          The full name of the implementing class for the DMO.
    * @param   colName
    *          The column name which needs to have a computed version. This is
    *          not the computed column name, but the name of its "parent".
    * @param   ignoreCase
    *          <code>true</code> if this is a character field which has the
    *          "ignore case" flag set.
    *          
    * @return  The DDL to alter the computed column, so that its formula is
    *          specified. 
    */
   public String getAlterComputedColumnString(Configuration config,
                                              String implClassName, 
                                              String colName,
                                              boolean ignoreCase)
   {
      final PersistentClass pc = config.getClassMapping(implClassName);
      Table table = pc.getIdentityTable();
      Column computedCol = 
         table.getColumn(new Column(getComputedColumnPrefix() + colName));
      
      StringBuilder buf = new StringBuilder();
      if (ignoreCase)
      {
         buf.append("upper(");
      }
      buf.append("rtrim(");
      buf.append(colName);
      buf.append(", ' \\t\\n\\r')");
      if (ignoreCase)
      {
         buf.append(")");
      }
      
      String dataType = 
         computedCol.getSqlType(this, 
            new Mapping()
            {
               public String getIdentifierPropertyName(String className) 
               throws MappingException
               {
                  return pc.getIdentifierProperty().getName();
               }

               public Type getReferencedPropertyType(String className, String propertyName)
               throws MappingException
               {
                  return pc.getReferencedProperty(propertyName).getType();
               }

               public IdentifierGeneratorFactory getIdentifierGeneratorFactory()
               {
                  return null;
               }

               public Type getIdentifierType(String className)
               throws MappingException
               {
                  if (!pc.hasIdentifierProperty() ) 
                     return null;

                  return pc.getIdentifier().getType();
               }

               public Type getPropertyType(String className,
                                           String propertyName)
               throws MappingException
               {
                  Property prop = pc.getProperty(propertyName);
                  return prop.getType();
               }
            });
      
      // the syntax to set the formula for the computed column is:
      // alter table <table> alter column <column> <type> as <formula>
      return String.format("alter table %s alter column %s %s as %s;",
                           table.getQuotedName(this), 
                           computedCol.getQuotedName(this),
                           dataType,
                           buf.toString());
   }

   /**
    * Prepare the given metadata query parameter by making it uppercase.
    * 
    * @param   param
    *          The metadata query parameter.
    *          
    * @return  the uppercased parameter.
    */
   public String prepareMetadataParameter(String param)
   {
      return param == null ? null : param.toUpperCase();
   }

   /**
    * Process the given metadata query result, by making it lowercase.
    * 
    * @param   value
    *          The metadata query result.
    *          
    * @return  the lowercased result.
    */
   public String processMetadataResult(String value)
   {
      return value == null ? null : value.toLowerCase();
   }

   /**
    * Check if the given value is a desceding clause for a column index.
    * 
    * @param   value
    *          The value to be checked.
    *          
    * @return  <code>true</code> if the value is the "D" or "d" string.
    */
   public boolean isMetadataSortDesc(String value)
   {
      return "D".equalsIgnoreCase(value);
   }
   
   /**
    * Check if the given value is an asceding clause for a column index.
    * 
    * @param   value
    *          The value to be checked.
    *          
    * @return  <code>true</code> if the value is the "A" or "a" string.
    */
   public boolean isMetadataSortAsc(String value)
   {
      return "A".equalsIgnoreCase(value);
   }

   /**
    * Get a list of database prepare DDL statements.
    * 
    * @param    database
    *           The database which needs to be prepared.
    *            
    * @return   The statement list.
    */
   public List<String> getDatabasePrepareStatements(Database database)
   {
      return H2Helper.getPrepareStatements(database);
   }
   
   /**
    * Check if the dialect needs to set an explicit collation during schema
    * DDL generation.
    * 
    * @return  always <code>true</code>.
    */
   public boolean explicitSetCollation()
   {
      return true;
   }
   
   /**
    * Get the statement delimiter for this dialect.
    * 
    * @return  the statement delimiter.
    */
   public String getDelimiter()
   {
      return ";";
   }
   
   /**
    * Set any database parameters before, to be available when the database
    * driver is loaded.
    * <p>
    * The "h2.sortNullsHigh" setting was removed, as it was determined that
    * by forcing the H2 to sort nulls high, it causes a regression in a test
    * where 4GL seems to sort the unknown value at the end of the result set,
    * when an descending sort clause was used (instead of sorting it at the
    * begining of the result set).
    * <p>
    * TODO: more testing needs to be done to determine exactly how 4GL sorts
    * the unknown value in the temporary database (and permanent DB too). At
    * the time of this comment, H2 has a bug which doesn't enforce the "sort
    * nulls high" setting in cases when the query's result order is determined
    * by the index.
    */
   public void beforeDriverLoad()
   {
       // System.setProperty("h2.sortNullsHigh", "true");
   }

   /**
    * Build the URL to be used for a remote connection, based on the given
    * URL and database.
    * 
    * @param    url
    *           The URL as received from the remote server.
    * @param    database
    *           The database to which this URL belongs.
    * 
    * @return   The built URL as to be used for a remote connection.
    *
    * @throws   PersistenceException
    *           if the specified connection URL is invalid.
    */
   public String buildRemoteURL(String url, Database database)
   throws PersistenceException
   {
      // save the original URL
      String origUrl = url;

      url = url.substring("jdbc:".length());
      
      // For a H2 database, the URL can't target an embedded database. So,
      // the "h2:tcp" or "h2:ssl" string must exist after the "jdbc:" 
      // prefix.
      // As with version 1.4.127, the H2 url format is:
      // { {.|mem:}[name] | [file:]fileName | 
      // {tcp|ssl}:[//]server[:port][,server2[:port]]/name }[;key=value...]
      
      // check the url's validity - to be a tcp or ssl-served url
      if (url.indexOf("h2:tcp:") != 0 && url.indexOf("h2:ssl:") != 0)
      {
         throw new PersistenceException(
               "URL \"" + origUrl + "\" is invalid: remote connections can " +
               " be acquired only for the 'tcp' or 'ssl' modes !");
      }

      try
      {
         // save the connection mode
         String mode = url.substring(0, "h2:tcp:".length());
         // get rid of the "h2:tcp:" or "h2:ssl:" text
         url = url.substring("h2:tcp:".length());
         // put back the "h2:" prefix
         url = "h2:" + url;
         
         URI uri = (new URI(url)).normalize();

         // Check for localhost references.
         String host = uri.getHost().toLowerCase();
         if (InetAddress.getByName(host).isLoopbackAddress())
         {
            // Use hostname stored in database object.
            host = database.getSocketAddress().getHostName();
            
            // Generate new URI.
            uri = new URI(uri.getScheme(),
                          uri.getUserInfo(),
                          host,
                          uri.getPort(),
                          uri.getPath(),
                          uri.getQuery(),
                          uri.getFragment());
            
            // Prepend 'jdbc:' prefix.
            StringBuilder buf = new StringBuilder("jdbc:");
            buf.append(mode);
            buf.append(uri.toString().substring("h2:".length()));
            url = buf.toString();
         }
         else
         {
            // no preparation is needed
            url = origUrl;
         }
      }
      catch (Exception exc)
      {
         throw new PersistenceException(
            "Invalid database connection URL", exc);
      }

      return url;
   }
}
