/*
** Module   : P2JPostgreSQLDialect.java
** Abstract : Implementation of P2JDialect for a PostgreSQL backend
**
** Copyright (c) 2004-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20060929   @30034 Created initial version. Adds P2JDialect
**                           method implementations to extend Hibernate's
**                           PostgreSQLDialect.
** 002 ECF 20070514   @33483 Added extractColumnName(). Extracts a root
**                           column name from its native representation in
**                           a particular dialect.
** 003 ECF 20070923   @35221 Override superclass' VARCHAR mapping to point
**                           to PG's 'text' data type. This is desirable
**                           because we always want to use 'text' instead
**                           of 'varchar(n)' for string data on the
**                           backend.
** 004 ECF 20071002   @35334 Implemented new P2JDialect methods. Added
**                           createContext() and postTransaction() methods
**                           to enable temp table periodic vacuums.
** 005 ECF 20080311   @37518 Implemented methods for computed columns and
**                           user defined function overloading, which are
**                           not needed by this dialect but required by
**                           the P2JDialect interface.
** 006 ECF 20080430   @38179 Fix to permit schema generation. Defer lookup
**                           of vacuumInterval in directory until first
**                           use. Performing the directory bind during
**                           class initialization fails when generating
**                           schema via Ant task.
** 007 ECF 20080611   @38705 Added isCaseInsensitiveColumn(). Provides a
**                           rough test for case-insensitivity, based only
**                           on an expression which contains a column
**                           name, such as might be extracted from index
**                           metadata.
** 008 ECF 20080805   @39316 Added getCreateSequenceString(). Generates
**                           DDL to create a sequence starting at a
**                           particular value.
** 009 CA  20080815   @39444 Added isConnectionError method - it checks if
**                           a given exception represents a connection 
**                           error.
** 010 ECF 20080930   @40001 Added isQueryRangeParameterInlined().
**                           Indicates whether query substitution
**                           parameters within range checks should be
**                           inlined.
** 011 ECF 20081015   @40096 Added formatDate(). Formats a date into a
**                           string suitable for use in an SQL statement.
** 012 ECF 20090310   @41456 Fixed performance problem with variable limit
**                           clause in a SELECT statement. In cases of
**                           simple and very small (2 or less) record
**                           limits (with no offset) in a SELECT, we now
**                           inline the number directly in the SQL. In the
**                           default implementation, Hibernate would use a
**                           bound parameter, which could result in a poor
**                           query plan (sequential table scan instead of
**                           an index scan).
** 013 GES 20090424   @41944 Import change.
** 014 ECF 20090705   @43225 Appended 'analyze' to 'vacuum' command.
** 015 SVL 20090810   @43575 Added requiresExplicitCastInsideTernary().
**                           Determines whether substitution parameters should
**                           be explicitly casted to their datatypes inside
**                           "then ? else ?" statement.
** 016 GES 20100701          Added a type name override to allow honoring of
**                           precision and scale in numeric types.
** 017 CA  20101026          Added methods to generate the DROP/CREATE INDEX
**                           statements in a dialect specific form. Also, 
**                           added method to generate the ALTER TABLE statement
**                           which sets the formula for the computed columns.
**                           Added methods to prepare/process string metadata 
**                           query parameters and results, to get the database
**                           prepare DDL statements and collation statement.
** 018 CA  20101202          Added method to build the URL for a remote 
**                           connection.
** 019 SVL 20110622          Added usesSingleBackslash.
** 020 OM  20121120          Implemented DDL/SQL generation for sequences.
** 021 SVL 20130331          Upgraded to Hibernate 4.
** 022 ECF 20130510          Modified sequence-related methods to not use quoted names. Replace
**                           deprecated superclass.
** 023 OM  20130712          Overrided default (invalid) min/max values for sequences.
*/

package com.goldencode.p2j.persist.dialect;

import java.net.InetAddress;
import java.net.URI;
import java.sql.*;
import java.sql.Date;
import java.text.*;
import java.util.*;

import org.apache.commons.logging.*;
import org.hibernate.cfg.Configuration;
import org.hibernate.dialect.*;
import org.hibernate.mapping.Column;
import org.hibernate.mapping.Index;
import org.hibernate.mapping.Table;

import com.goldencode.util.*;
import com.goldencode.p2j.directory.DirectoryService;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.security.ContextLocal;
import com.goldencode.p2j.util.*;

/**
 * Extends Hibernate's <code>PostgreSQLDialect</code> class to add method
 * implementations required by the {@link P2JDialect} interface.
 * <p>
 * Note:  this implementation expects a directory to be available at the time
 * the class is loaded.
 */
public class P2JPostgreSQLDialect
extends PostgreSQL82Dialect
implements P2JDialect
{
   /** Logger */
   private static final Log LOG = LogFactory.getLog(P2JPostgreSQLDialect.class);
   
   /** Default minimum interval in seconds between temp table vacuums */
   private static final int DEFAULT_VACUUM_INTERVAL = 30;
   
   /** Debug level logging enabled? */
   private static final boolean debug = LOG.isDebugEnabled();
   
   /** Flag indicating whether to inline a record limit in an SQL SELECT */
   private static final ContextLocal<Boolean> inlineLimit =
   new ContextLocal<Boolean>()
   {
      protected Boolean initialValue() { return Boolean.FALSE; }
   };
   
   /** Lowest date supported by PostgreSQL (01/01/4713 BC) */
   private static final Date lowDate;
   
   /** Highest date supported by PostgreSQL (12/31/32767 AD) */
   private static final Date highDate;
   
   /** Actual minimum interval in milliseconds between temp table vacuums */
   private static long vacuumInterval = -1L;
   
   static
   {
      // Prepare supported date values.  We use the GMT timezone to define
      // our date range, since this is how com.goldencode.p2j.util.date
      // handles its dates.
      
      TimeZone tz = TimeZone.getTimeZone("GMT");
      GregorianCalendar cal = new GregorianCalendar(tz);
      
      cal.set(Calendar.ERA, GregorianCalendar.BC);
      cal.set(Calendar.YEAR, 4713);
      cal.set(Calendar.MONTH, Calendar.JANUARY);
      cal.set(Calendar.DATE, 1);
      lowDate = new Date(cal.getTime().getTime());
      
      cal.clear();
      
      cal.set(Calendar.ERA, GregorianCalendar.AD);
      cal.set(Calendar.YEAR, 32767);
      cal.set(Calendar.MONTH, Calendar.DECEMBER);
      cal.set(Calendar.DATE, 31);
      highDate = new Date(cal.getTime().getTime());
   }
   
   /**
    * Default constructor.  This constructor is for Hibernate's use.  Use code
    * should not construct instances of this class.
    * <p>
    * We provide this constructor to override the superclass' mapping of 
    * <code>Types.VARCHAR</code>, in order to force this type to map to
    * PostgreSQL's <code>text</code> data type.
    */
   public P2JPostgreSQLDialect()
   {
      super();
      
      // Override superclass' VARCHAR mapping, since we always use text for
      // string data.
      registerColumnType(Types.VARCHAR, "text");
      
      // Override superclass' NUMERIC mapping, since the Hibernate PostgreSQL
      // dialect doesn't honor precision and scale.
      registerColumnType(Types.NUMERIC, "numeric($p,$s)");
   }
   
   /**
    * Report the minimum number of milliseconds between explicit vacuums.
    * This number is retrieved from the directory, or if not configured there,
    * is defaulted to a standard interval.
    * 
    * @return  Minimum number of milliseconds between vacuums.
    */
   private static long getVacuumInterval()
   {
      if (vacuumInterval < 0)
      {
         // Attempt to read temp-table vacuum interval from temp-table section
         // of directory.  Missing value uses default interval, 0 disables,
         // negative number is an error.
         DirectoryService ds = DirectoryService.getInstance();
         if (!ds.bind())
         {
            throw new RuntimeException("Directory bind failed");
         }
         
         try
         {
            String nodeID = Utils.findDirectoryNodePath(ds,
                                                        "database", 
                                                        "container",
                                                        false);
            Integer value = null;
            if (nodeID != null)
            {
               StringBuilder buf = new StringBuilder(nodeID);
               buf.append("/");
               buf.append(DatabaseManager.TEMP_TABLE_DB.getName());
               buf.append("/p2j/vacuum_interval");
               value = ds.getNodeInteger(buf.toString(), "value");
            }
            
            boolean info = LOG.isInfoEnabled();
            int seconds = 0;
            if (value == null)
            {
               seconds = DEFAULT_VACUUM_INTERVAL;
               if (info)
               {
                  LOG.info(
                     "Temp table vacuum interval not specified;  using default");
               }
            }
            else
            {
               seconds = value.intValue();
            }
            
            if (seconds <= 0)
            {   
               if (LOG.isWarnEnabled())
               {
                  LOG.warn("Temp table periodic vacuuming disabled!");
               }
            }
            else if (LOG.isInfoEnabled())
            {
               LOG.info("Temp table vacuum interval set to " +
                        seconds +
                        " seconds");
            }
            
            vacuumInterval = seconds * 1000;
         }
         finally
         {
            ds.unbind();
         }
      }
      
      return vacuumInterval;
   }
   
   /**
    * Indicate whether this dialect requires computed columns to participate
    * within index definitions, instead of embedded expressions.
    * 
    * @return  <code>false</code> to indicate the backing database can embed
    *          an expression within an index definition.
    */
   public boolean needsComputedColumns()
   {
      return false;
   }
   
   /**
    * Get the prefix, if any, which should be prepended to computed column
    * and/or property names.
    * 
    * @return  <code>null</code> to indicate the backing database does not use
    *          computed columns.
    * 
    * @see     #needsComputedColumns()
    */
   public String getComputedColumnPrefix()
   {
      return null;
   }

   /**
    * Indicate whether this dialect natively supports the overloading of user
    * defined functions by parameter number and data type.
    * 
    * @return  <code>true</code>.
    */
   public boolean supportsFunctionOverloading()
   {
      return true;
   }
   
   /**
    * Create a data object, the purpose of which is specific to a particular
    * dialect implementation, to be stored as a context-local object.  This
    * allows a dialect to establish context-local data which is stored by the
    * persistence runtime framework, without the overhead of a separate,
    * context-local variable.  This object will be handed back to the dialect
    * in callback methods when the services of certain hooks are required.
    * <p>
    * Since the context-local data is specific to a particular dialect
    * implementation, it is optional.  Dialect implementations which do not
    * have a need for this data should simply return <code>null</code>.
    * 
    * @param   persistence
    *          Provider of persistence services.
    * 
    * @return  Dialect-specific, context-local data, or <code>null</code>.
    */
   public Object createContext(Persistence persistence)
   {
      Database database = persistence.getDatabase();
      
      return (DatabaseManager.TEMP_TABLE_DB.equals(database) &&
              getVacuumInterval() > 0
              ? (new Context())
              : null);
   }
   
   /**
    * A dialect-specific callback hook which is invoked by the persistence
    * runtime framework after a transaction is complete.  The Hibernate
    * session will have been closed by the time this method is invoked.
    * 
    * @param   persistence
    *          Provider of persistence services.
    * @param   context
    *          Optional, context-local data created previously by {@link
    *          #createContext}.
    * @param   rollback
    *          <code>true</code> if transaction ended with a rollback;
    *          <code>false</code> if transaction ended with a commit.
    * 
    * @throws  PersistenceException
    *          if any persistence-related error occurs.
    */
   public void postTransaction(Persistence persistence,
                               Object context,
                               boolean rollback)
   throws PersistenceException
   {
      if (context != null)
      {
         Context local = (Context) context;
         long now = System.currentTimeMillis();
         long diff = now - local.lastVacuum;
         if (diff >= getVacuumInterval())
         {
            int count = 0;
            try
            {
               StringBuilder buf = new StringBuilder();
               Iterator<String> iter = TemporaryBuffer.activeTables();
               for (count = 0; iter.hasNext(); count++, buf.setLength(0))
               {
                  buf.append("vacuum analyze ");
                  buf.append(iter.next());
                  persistence.executeSQL(buf.toString());
               }
            }
            finally
            {
               long end = System.currentTimeMillis();
               long elapsed =  end - now;
               local.vacuumTime += elapsed;
               if (debug)
               {
                  StringBuilder buf = new StringBuilder("Vacuum ");
                  buf.append(count);
                  buf.append(" temp table(s) [");
                  buf.append(diff);
                  buf.append(":");
                  buf.append(elapsed);
                  buf.append(":");
                  buf.append(local.vacuumTime);
                  buf.append("]");
                  LOG.debug(persistence.message(buf.toString()));
               }
               local.lastVacuum = end;
            }
         }
      }
   }
   
   /**
    * Retrieve the earliest possible date value which is valid for the
    * backing database.
    *
    * @return  Earliest date supported by the backing database.
    */
   public Date getLowDate()
   {
      return lowDate;
   }
   
   /**
    * Retrieve the latest possible date value which is valid for the backing
    * database.
    *
    * @return  Latest date supported by the backing database.
    */
   public Date getHighDate()
   {
      return highDate;
   }
   
   /**
    * Format a date object as a string which is appropriate for use in an SQL
    * statement.
    * 
    * @param   d
    *          Date to be formatted as a string.
    * 
    * @return  Formatted representation of the given date.
    */
   public String formatDate(date d)
   {
      NumberFormat nf = NumberFormat.getInstance(); 
      
      nf.setMinimumIntegerDigits(4);
      nf.setGroupingUsed(false);
      
      StringBuffer buf = new StringBuffer();
      int year = d.getYear();
      
      buf.append(nf.format(Math.abs(year)));
      buf.append('-');
      
      nf.setMinimumIntegerDigits(2);
      buf.append(nf.format(d.getMonth()));
      buf.append('-');
      buf.append(nf.format(d.getDay()));
      
      if (year < 0)
      {
         buf.append(" BC");
      }
      
      return buf.toString();
   }
   
   /**
    * Given a column expression as queried from database metadata, extract the
    * root name of a column.  For example, a column name as found within an
    * index in metadata might be represented as:
    * <pre>
    *    upper(rtrim("key"::text, ' \t\n\r'::text))
    * <pre>
    * We are interested only in the <code>key</code> root name, so we strip
    * away the upper() and rtrim() functions, the double quotes, and the
    * <code>::text</code> suffix.
    *
    * @param   expr
    *          Expression containing column name as natively represented by
    *          the database dialect.
    *
    * @return  Root column name.
    */
   public String extractColumnName(String expr)
   {
      final String rtrim = "rtrim(";
      final String upper = "upper(";
      
      String root = expr.toLowerCase();
      
      // Remove "upper" function if present (only applicable to text fields).
      if (root.startsWith(upper))
      {
         int len = root.length();
         root = root.substring(upper.length(), len - 1);
      }
      
      // Remove "rtrim" function if present (only applicable to text fields).
      if (root.startsWith(rtrim))
      {
         int commaPos = root.lastIndexOf(",");
         root = root.substring(rtrim.length(), commaPos);
      }
      
      // Remove data type suffix.
      int delim = root.indexOf("::");
      
      root = (delim >= 0 ? root.substring(0, delim) : root);
      
      // Strip enclosing parentheses.
      while (root.startsWith("(") && root.endsWith(")"))
      {
         root = root.substring(1, root.length() - 1);
      }
      
      // Strip enclosing double quotes or backticks.
      root = StringHelper.stripEnclosing(root, '"');
      root = StringHelper.stripEnclosing(root, '`');
      
      return root;
   }
   
   /**
    * Adjust a character column name appropriate to this dialect, such that it
    * can be inserted into an index.  This involves wrapping the name within
    * an rtrim function, and if <code>ignoreCase</code> is <code>true</code>,
    * an upper function.
    * 
    * @param   name
    *          Base column name.
    * @param   ignoreCase
    *          <code>true</code> to cause case to be ignored, else
    *          <code>false</code>.
    * 
    * @return  Processed column name.
    */
   public String getProcessedCharacterColumnName(String name, boolean ignoreCase)
   {
      StringBuilder buf = new StringBuilder();
      if (ignoreCase)
      {
         buf.append("upper(");
      }
      buf.append("rtrim(");
      buf.append(name);
      buf.append(", ' \t\n\r')");
      if (ignoreCase)
      {
         buf.append(")");
      }
      
      return buf.toString();
   }
   
   /**
    * Report whether the given column name represents an alias or an
    * expression which indicates the column is case insensitive.  Assume the
    * column represented by <code>name</code> is already determined to be the
    * appropriate data type for textual data.
    * 
    * @param   name
    *          Column name, alias, or expression.
    * 
    * @return  <code>true</code> if <code>name</code> represents
    *          case-insensitive data.
    */
   public boolean isCaseInsensitiveColumn(String name)
   {
      return name.toLowerCase().indexOf("upper(") >= 0;
   }
   
   /**
    * Generate a DDL string which will create a global sequence with the given
    * starting value.
    * 
    * @param   name
    *          Sequence name.
    * @param   start
    *          Sequence starting value.
    * 
    * @return  Dialect-specific string to create the sequence, or
    *          <code>null</code> if not supported.
    */
   public String getCreateSequenceString(String name, long start)
   {
      StringBuilder buf = new StringBuilder(getCreateSequenceString(name));
      buf.append(" start ");
      buf.append(start);
      
      return buf.toString();
   }

   /**
    * Generate a DDL string which will create a sequence with given properties.
    * 
    * @param   name
    *          Sequence name. Mandatory not null.
    * @param   init
    *          Sequence starting value. Mandatory(not null).
    * @param   increment
    *          Sequence incrementing value. Mandatory(not null).
    * @param   minVal
    *          Sequence minimum value. Optional (may be null).
    * @param   maxVal
    *          Sequence maximum value. Optional (may be null).
    * @param   cycle
    *          Sequence cycling property. Optional, if missing assumed not cycling.
    * 
    * @return  Dialect-specific string to create the sequence, or <code>null</code> if 
    *          not supported.
    */
   @Override
   public String getCreateSequenceString(String name,
                                         Long init,
                                         Long increment,
                                         Long minVal,
                                         Long maxVal,
                                         Boolean cycle)
   {
      // in Progress a cycling sequence restarts at initial value instead of min/max
      // in PSQL, such sequence will warp to other limit of the sequence.
      if (cycle && increment > 0)
      {
         minVal = init;
      }
      else
      {
         maxVal = init;
      }
      
      
      StringBuilder buf = new StringBuilder(getCreateSequenceString(name));
      buf.append(" increment ").append(increment);
      
      if (minVal != null)
      {
         buf.append(" minvalue ").append(minVal);
      }
      else 
      {
         // default MINVALUE in PSQL is 1, must take care to avoid 
         // START value (n) cannot be less than MINVALUE (1) error
         if (increment > 0 && init < 1)
         {
            buf.append(" minvalue ").append(init);
         }
      }
      
      if (maxVal != null)
      {
         buf.append(" maxvalue ").append(maxVal);
      }
      else 
      {
         // default MAXVALUE in PSQL is -1, must take care to avoid 
         // START value (n) cannot be greater than MAXVALUE (-1) error
         if (increment < 0 && init > -1)
         {
            buf.append(" maxvalue ").append(init);
         }
      }
      
      buf.append(" start ").append(init);
      if (cycle != null && cycle)
      {
         buf.append(" cycle");
      }
      buf.append(" cache 1");

      return buf.toString();
   }

   /**
    * Generate a DDL string which will drop a sequence with given properties.
    * 
    * @param   name
    *          The name of the sequence to be dropped.
    *
    * @return  Dialect-specific string to drop the sequence, or
    *          <code>null</code> if not supported.
    */
   @Override
   public String getDropSequenceString(String name)
   {
      StringBuilder buf = new StringBuilder("drop sequence if exists ");
      buf.append(name);

      return buf.toString();
   }

   /**
    * Generate the appropriate select statement to retrieve the next value of a sequence.
    * 
    * @param   sequenceName
    *          The name of the sequence.
    * @param   newVal
    *          The new value to be set.
    *
    * @return  Dialect-specific string to query the sequence, or
    *          <code>null</code> if not supported.
    */
   @Override
   public String getSequenceSetValString(String sequenceName, long newVal)
   {
      StringBuilder sb = new StringBuilder("select setval('");
      sb.append(sequenceName).append("', ");
      sb.append(newVal).append(")");
      
      return sb.toString();
   }

   /**
    * Generate the appropriate select statement to reset the value of a sequence.
    * 
    * @param   sequenceName
    *          The name of the sequence to be queried.
    *
    * @return  Dialect-specific string to query the sequence, or
    *          <code>null</code> if not supported.
    */
   @Override
   public String getSequenceCurrValString(String sequenceName)
   {
      // The CURRVAL is relative to current session and will fail if called
      // before a call to NEXTVAL or SETVAL
      StringBuilder sb = new StringBuilder();
      sb.append("select last_value from ");
      sb.append(sequenceName);
      
      return sb.toString();
   }

   /**
    * Returns <code>true</code> if this dialect fully support the sequences as
    * 4GL language (most important: bounded values, cycle). If not fully
    * supported, additional code is needed in P2J sequence class
    * implementation to fix this.
    * 
    * @return true because PostgreSQL dialect fully support 4GL sequences
    */
   @Override
   public boolean hasFullSequenceSupport()
   {
      return true;
   }

   /**
    * Context local data needed by this object to perform dialect-specific
    * processing.  In this case, we track the last time a temp table vacuum
    * was performed and how much accumulated time we have spent performing
    * such vacuums in this context.
    */
   private static class Context
   {
      /** Time of last temp table vacuum */
      private long lastVacuum = System.currentTimeMillis();
      
      /** Accumulated time spent vacuuming temp tables */
      private long vacuumTime = 0L;
   }

   /**
    * Given an exception, this method will check if it indicates a database
    * connectivity error.
    * <p>
    * For the PostgreSQL dialect, if the state indicates a 08xxx error code
    * or 57P01 or 57P02 codes, the method will return <code>true</code>.
    * <p>
    * A complete list of defined error codes can be found here:
    * http://www.postgresql.org/docs/8.1/static/errcodes-appendix.html
    * 
    * @param   exc
    *          The exception to be checked.
    * 
    * @return  <code>true</code> if the exception indicates a database 
    *          connectivity error.
    */
   public boolean isConnectionError(Exception exc)
   {
      boolean error = false;
      
      Throwable next = exc;
      while (next != null && !error)
      {
         if (next instanceof SQLException)
         {
            String state = ((SQLException) next).getSQLState();

            error = state != null && 
                    (state.startsWith("08") || 
                     state.equals("57P01")  ||
                     state.equals("57P02"));
         }
         next = next.getCause();
      }
      
      return error;
   }

   /**
    * Indicate whether a query parameter which is part of a range check should
    * be inlined into the where clause, or whether it should be substituted
    * into a prepared statement.  This will result in an "inlined"
    * substitution parameter within a where clause.  For example, given a
    * where clause of:
    * <pre>
    *    where value >= ? and id = ?
    * </pre>
    * along with respective substitution parameters of 45 and 100, the
    * inlined version would read:
    * <pre>
    *    where value >= 45 and id = ?
    * </pre>
    * Only the <code>id</code> substitution parameter of 100 would be
    * substituted into the prepared statement.
    * <p>
    * Inlining in such a way can give some database dialects the opportunity
    * to choose a better query plan, than if the parameter being tested with
    * a range check were to be substituted at statement execution time.
    * <p>
    * Note:  this serves as a hint which may be overridden by runtime logic or
    * other configuration settings.
    * 
    * @return  <code>true</code> to inline range check parameters;
    *          <code>false</code> to use normal parameter substitution.
    */
   public boolean isQueryRangeParameterInlined()
   {
      return true;
   }
   
   /**
    * Enable/disable limit clause inlining.  Inlining the limit clause will
    * cause the limit/max value to be hard-coded into the SQL statement,
    * rather than being bound as a substitution parameter.
    * <p>
    * Dialects are free to treat this method as a hint.
    * 
    * @param   on
    *          <code>true</code> to enable inlining;  <code>false</code> to
    *          disable it.
    */
   public void inlineLimit(boolean on)
   {
      inlineLimit.set(on);
   }

   /**
    * Determines whether substitution parameters should be explicitly casted
    * to their datatypes inside "then ? else ?" statement.
    *
    * @return <code>false</code>.
    */
   public boolean requiresExplicitCastInsideTernary()
   {
      return true;
   }

   /**
    * Determines whether dialect uses single backslash (e.g. '12\34') into
    * string literals rather than double backslash (e.g. '12\\34').
    *
    * @return <code>false</code>.
    */
   public boolean usesSingleBackslash()
   {
      return false;
   }

   /**
    * Override the parent's implementation to inline small (2 or less) limit
    * values when there is no accompanying offset.  In other cases, we fall
    * back to the original implementation.
    * <p>
    * WARNING:  the implementation of this override has an ugly dependency on
    * the order of the logic in Hibernate's <code>Loader</code> class.
    * Specifically, we rely on the fact that this method is invoked when a
    * query is being prepared, which must happen before parameters are bound,
    * at which point {@link #supportsVariableLimit()} is invoked.  This
    * method sets state in a context local variable which is later read when
    * the latter method is invoked.
    * <p>
    * TODO:  review this implementation when upgrading Hibernate versions. 
    * 
    * @param   sql
    *          Base select statement SQL without the limit/offset clause.
    * @param   offset
    *          Starting offset.
    * @param   limit
    *          Number of records to return.
    *          
    * @return  SQL with limit/offset clause appended.
    * 
    * @see     #supportsVariableLimit()
    */
   public String getLimitString(String sql, int offset, int limit)
   {
      String stmt;
      boolean inlinePermitted = inlineLimit.get();
      
      if (inlinePermitted && limit <= 2 && offset == 0)
      {
         StringBuilder buf = new StringBuilder(sql);
         buf.append(" limit ");
         buf.append(limit);
         stmt = buf.toString();
      }
      else
      {
         if (inlinePermitted)
         {
            // Override inline request.
            inlineLimit.set(Boolean.FALSE);
         }
         stmt = super.getLimitString(sql, offset, limit);
      }
      
      return stmt;
   }
   
   /**
    * Indicate whether a variable limit is supported by this dialect.  The
    * answer is dependent upon context local state set within {@link
    * #getLimitString(String, int, int)}.  See the applicable warning in that
    * method's description.  This method is consulted during parameter binding
    * in Hibernate's <code>Loader</code> class.
    * 
    * @return  <code>true</code> if a bound parameter should be used in a
    *          SELECT statement's limit clause;  <code>false</code> if the
    *          limit was inlined by {@link #getLimitString(String, int, int)}.
    */
   public boolean supportsVariableLimit()
   {
      return (inlineLimit.get() ? false : super.supportsVariableLimit());
   }

   /**
    * Build the DDL to drop an index based upon the specified definition.
    * The SQL is generated by using the database-specific dialect.
    *
    * @param   index
    *          Definition of the index to be dropped.
    *
    * @return  The DDL to drop the given index.
   */
   public String getDropIndexString(P2JIndex index)
   {
      return "drop index " + index.getName();
   }

   /**
    * Build the DDL to create an index based upon the specified definition.
    * The SQL is generated by using a database-specific dialect.
    *
    * @param   index
    *          Definition of the index to be created.
    * @param   unique
    *          If <code>true</code>, create the index as a unique index,
    *          <code>false</code> to omit the unique constraint.
    * @param   toFile
    *          <code>true</code> if the string will be output to a file or
    *          <code>false</code> if it will be directly submitted via JDBC.
    *
    * @return  The DDL which was created.
   */
   public String getCreateIndexString(P2JIndex index, 
                                      boolean  unique, 
                                      boolean  toFile)
   {
      // Add primary key column to any index which is not explicitly or
      // implicitly unique.
      // TODO:  move this to import ruleset so we don't hardcode primary
      // key column name.
      if (!index.isUnique())
      {
         index.addComponent("id");
      }
      
      // Extract index information into a Hibernate-friendly form.
      String tabName = index.getTable();
      Table table = new Table(tabName);
      List<Column> columns = new ArrayList<Column>();
      Iterator<P2JIndexComponent> iter = index.components();
      while (iter.hasNext())
      {
         P2JIndexComponent next = iter.next();
         String colName = next.getName();
         if (next.isCharType())
         {
            // Special handling for text columns.
            boolean ignoreCase = next.isIgnoreCase();
            StringBuilder buf = new StringBuilder();
            if (ignoreCase)
            {
               buf.append("upper(");
            }
            buf.append("rtrim(");
            buf.append(colName);
            buf.append(", ");
            
            if (toFile)
            {
               // file output with PostgreSQL needs different syntax to
               // annotate that this is a string literal
               buf.append("E");
            }
            
            buf.append("' \t\n\r')");
            if (ignoreCase)
            {
               buf.append(")");
            }
            colName = buf.toString();
         }
         
         columns.add(new Column(colName));
      }
      
      // TODO: asc/desc is not supported by hibernate
      // current version of PostgreqSQL - 8.2.x doesn't support "asc/desc"
      // clause, only from 8.3.x and up supports it
      
      // Compose the string.
      String createString =
         Index.buildSqlCreateIndexString(this,
                                         index.getName(),
                                         table,
                                         columns.iterator(),
                                         unique,
                                         null,
                                         null);
      
      if (toFile)
      {
         createString = createString.replaceAll("\t\n\r", "\\\\t\\\\n\\\\r");
      }
      
      return createString;
   }
   
   /**
    * The PostgreSQL dialect doesn't need computed columns for indexed, char
    * columns, so this implementation returns <code>null</code>
    * 
    * @param   config
    *          The hibernate's configuration object.
    * @param   implClassName
    *          The full name of the implementing class for the DMO.
    * @param   colName
    *          The column name which needs to have a computed version. This is
    *          not the computed column name, but the name of its "parent".
    * @param   ignoreCase
    *          <code>true</code> if this is a character field which has the
    *          "ignore case" flag set.
    *          
    * @return  <code>null</code> 
    */
   public String getAlterComputedColumnString(Configuration config,
                                              String implClassName, 
                                              String colName,
                                              boolean ignoreCase)
   {
      return null;
   }

   /**
    * Prepare the given metadata query parameter by making it lowercase.
    * 
    * @param   param
    *          The metadata query parameter.
    *          
    * @return  the lowercased parameter.
    */
   public String prepareMetadataParameter(String param)
   {
      return param == null ? null : param.toLowerCase();
   }

   /**
    * Process the given metadata query result, by making it lowercase.
    * 
    * @param   value
    *          The metadata query result.
    *          
    * @return  the lowercased result.
    */
   public String processMetadataResult(String value)
   {
      return value == null ? null : value.toLowerCase();
   }

   /**
    * Check if the given value is a desceding clause for a column index.
    * 
    * @param   value
    *          The value to be checked.
    *          
    * @return  <code>true</code> if the value is the "D" or "d" string.
    */
   public boolean isMetadataSortDesc(String value)
   {
      return "D".equalsIgnoreCase(value);
   }
   
   /**
    * Check if the given value is an asceding clause for a column index.
    * 
    * @param   value
    *          The value to be checked.
    *          
    * @return  <code>true</code> if the value is the "A" or "a" string.
    */
   public boolean isMetadataSortAsc(String value)
   {
      return "A".equalsIgnoreCase(value);
   }

   /**
    * Get a list of database prepare DDL statements.
    * 
    * @param    database
    *           The database which needs to be prepared.
    *            
    * @return   an empty list.
    */
   public List<String> getDatabasePrepareStatements(Database database)
   {
      return Collections.EMPTY_LIST;
   }
   
   /**
    * Check if the dialect needs to set an explicit collation during schema
    * DDL generation.
    * 
    * @return  always <code>false</code>.
    */
   public boolean explicitSetCollation()
   {
      return false;
   }
   
   /**
    * Get the statement delimiter for this dialect.
    * 
    * @return  the statement delimiter.
    */
   public String getDelimiter()
   {
      return ";";
   }
   /**
    * Set any database parameters before, to be available when the database
    * driver is loaded.
    * <p>
    * This implementation does nothing.
    */
   public void beforeDriverLoad()
   {
      // no-op
   }

   /**
    * Build the URL to be used for a remote connection, based on the given
    * URL and database.
    * 
    * @param    url
    *           The URL as received from the remote server.
    * @param    database
    *           The database to which this URL belongs.
    * 
    * @return   The built URL as to be used for a remote connection.
    *
    * @throws   PersistenceException
    *           if the specified connection URL is invalid.
    */
   public String buildRemoteURL(String url, Database database)
   throws PersistenceException
   {
      String origUrl = url;
      try
      {
         url = url.substring("jdbc:".length());
         URI uri = (new URI(url)).normalize();

         // Check for localhost references.
         String host = uri.getHost().toLowerCase();
         if (InetAddress.getByName(host).isLoopbackAddress())
         {
            // Use hostname stored in database object.
            host = database.getSocketAddress().getHostName();
            
            // Generate new URI.
            uri = new URI(uri.getScheme(),
                          uri.getUserInfo(),
                          host,
                          uri.getPort(),
                          uri.getPath(),
                          uri.getQuery(),
                          uri.getFragment());
            
            // Prepend 'jdbc:' prefix.
            StringBuilder buf = new StringBuilder("jdbc:");
            buf.append(uri);
            url = buf.toString();
         }
         else
         {
            // no preparation is needed
            url = origUrl;
         }
      }
      catch (Exception exc)
      {
         throw new PersistenceException(
            "Invalid database connection URL", exc);
      }

      return url;
   }
}
