/*
** Module   : H2SequenceHandler.java
** Abstract : Static class that handles sequence functions and statements 
**            from Progress / OpenEdge 10.x.
**
** Copyright (c) 2012-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 OM  20121116 Created initial version with CURRVAL, NEXTVAL, and SETVAL support only.
** 002 SVL 20120331 Upgraded to Hibernate 4.
** 003 OM  20120712 Fixed extraction data from ResultSets. Added error handling.
**                  Fixed sequence list query from database, enforced cycle and bounds checking.
*/

package com.goldencode.p2j.persist.sequence;

import java.util.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.dialect.*;
import com.goldencode.p2j.util.*;
import org.hibernate.*;

/**
 * H2 implementation of sequences using a hybrid native SQL sequence primitives and 4GL fixes
 * backed by directory. The primary goal is to detect unknown (?) values that are not supported by
 * H2 and also full management of cycling sequences.
 * H2 only support incremental sequences, no cycling, no limits.
 */
public class H2SequenceHandler
extends SequenceHandler
{
   /**
    * The list of sequences grouped by databases.
    * It uses lazy initialization.
    */
   private HashMap<String, LinkedList<String>> sequences = null;

   /**
    * Queries the current value of a sequence.
    * @param   seqName
    *          The name of the sequence.
    * @param   ldbName
    *          The database to which the sequence belongs to. 
    *          It is assumed it is not null and valid.
    *
    * @return the current value of the sequence or unknown value (?) if
    *          sequence/database not found or the sequence is not cycling
    *          and has already passed the maximum value.
    */
   @Override
   public int64 getCurrentValue(String seqName, String ldbName)
   {
      Sequence seq = SequenceManager.getSequence(ldbName, seqName);
      if (seq == null)
      {
         ErrorManager.recordOrThrowError(2913,
                 "Unknown sequence expression " + seqName.toUpperCase());
         return new int64();
      }

      Persistence persistence = ConnectionManager.getPersistence(ldbName);
      P2JDialect dialect = (P2JDialect) persistence.getDialect();
      String strQuery = dialect.getSequenceCurrValString(seqName);
      long ret = 0;

      try
      {
         ScrollableResults rs = persistence.executeSQLQuery(strQuery);
         if (rs.next())
         {
            Object result = rs.get(0);
            if (result instanceof java.lang.Long)
            {
               // deprecated ?
               ret = (java.lang.Long) rs.get(0);
            }
            else if (result instanceof java.math.BigInteger)
            {
               System.out.println(result);
               ret = ((java.math.BigInteger) rs.get(0)).longValue();
            }
            else 
            {
               // don't know how to cast the result:
               System.err.println("** H2SequenceHandler.getCurrentValue() cannot handle " +
                     result.getClass() + "result type.");
               return new int64();
            }
         }
      }
      catch (PersistenceException pe)
      {
         pe.printStackTrace();
         return new int64(); // return unknown ?
      }

      // check the result against the dmo-index properties of the sequence
      long inc = seq.getIncrement();
      if (inc > 0)
      {
         // ascendant sequence
         if (ret > seq.getMax())
         {
            // current-value return the last valid value
            ret = seq.getMax();
         }
      }
      else if (inc < 0)
      {
         // descendant sequence
         if (ret < seq.getMin())
         {
            // current-value return the last valid value
            ret = seq.getMin();
         }
      }

      return new int64(ret);
   }

   /**
    * Computes and returns the next value of a sequence.
    * @param   seqName
    *          The name of the sequence.
    * @param   ldbName
    *          The database to which the sequence belongs to. 
    *          It is assumed it is not null and valid.
    *
    * @return the current value of the sequence or unknown value (?) if
    *          sequence/database not found or the sequence is not cycling
    *          and has passed the maximum/minimum value.
    */
   @Override
   public int64 getNextValue(String seqName, String ldbName)
   {
      Sequence seq = SequenceManager.getSequence(ldbName, seqName);
      if (seq == null)
      {
         ErrorManager.recordOrThrowError(2913,
                 "Unknown sequence expression " + seqName.toUpperCase());
         return new int64();
      }

      long crtVal = getCurrentValue(seqName, ldbName).longValue();
      long inc = seq.getIncrement();
      
      // H2 server is different than Progress, it has not MIN/MAX attributes: 
      //    * normally there is no CYCLE attribute
      //    * always warps at Long.MAX_VALUE and Long.MIN_VALUE 
      // We need to detect this occurrences and return ? instead.

      if (!seq.isCycle())
      {
         if (inc > 0)
         {
            if (crtVal + inc > seq.getMax() ||     // prevent H2 from passing MAX-value
                crtVal + inc < crtVal)             // prevent H2 from warping downwards on long
            {
               return new int64();
            }
         }
         else 
         {
            if (crtVal + inc < seq.getMin() ||     // prevent H2 from passing MIN-value
                crtVal + inc > crtVal)             // prevent H2 from warping upwards on long
              {
                 return new int64();
              }
         }
      }
      else 
      {
         if (inc > 0)
         {
            if (crtVal + inc > seq.getMax() ||     // prevent H2 from passing MAX-value
                crtVal + inc < crtVal)             // prevent H2 from warping downwards on long
            {
               long ret = seq.getInitial(); 
               setValue(seqName, ldbName, ret);
               return new int64(ret);
            }
         }
         else
         {
            if (crtVal + inc < seq.getMin() ||     // prevent H2 from passing MIN-value
                crtVal + inc > crtVal)             // prevent H2 from warping upwards on long
            {
               long ret = seq.getInitial(); 
               setValue(seqName, ldbName, ret);
               return new int64(ret);
            }
         }
      }
      
      Persistence persistence = ConnectionManager.getPersistence(ldbName);
      P2JDialect dialect = (P2JDialect) persistence.getDialect();
      String strQuery = dialect.getSequenceNextValString(seqName);
      long ret = 0;

      try
      {
         ScrollableResults rs = persistence.executeSQLQuery(strQuery);
         if (rs.next())
         {
            Object result = rs.get(0);
            if (result instanceof java.lang.Long)
            {
               // deprecated ?
               ret = (java.lang.Long) rs.get(0);
            }
            else if (result instanceof java.math.BigInteger)
            {
               System.out.println(result);
               ret = ((java.math.BigInteger) rs.get(0)).longValue();
            }
            else 
            {
               // don't know how to cast the result:
               System.err.println("** H2SequenceHandler.getNextValue() cannot handle " +
                     result.getClass() + "result type.");
               return new int64();
            }
         }
      }
      catch (PersistenceException pe)
      {
         pe.printStackTrace();
         return new int64(); // return unknown ?
      }

      // check the result against the dmo-index properties of the sequence
      if (inc > 0)
      {
         // ascendant sequence
         if (ret > seq.getMax())
         {
            if (seq.isCycle())
            {
               // compute the correct value by restart the counter
               ret = seq.getInitial();

               // also set to database:
               setValue(seqName, ldbName, ret);
            }
            else
            {
               // terminating sequence, returning unknown
               return new int64();
            }
         }
      }
      else if (inc < 0)
      {
         // descendant sequence
         if (ret < seq.getMin())
         {
            if (seq.isCycle())
            {
               // compute the correct value by restart the counter
               ret = seq.getMax();

               // also set to database:
               setValue(seqName, ldbName, ret);
            }
            else
            {
               // terminating sequence, returning unknown
               return new int64();
            }
         }
      }

      return new int64(ret);
   }

   /**
    * Initialize (reset) the current value of a sequence.
    * @param   seqName
    *          The name of the sequence.
    * @param   ldbName
    *          The database to which the sequence belongs to. 
    *          It is assumed it is not null and valid.
    * @param   newVal
    *          The value the sequence will be reseted to.
    */
   @Override
   public void setValue(String seqName, String ldbName, long newVal)
   {
      Sequence seq = SequenceManager.getSequence(ldbName, seqName);
      if (seq == null)
      {
         ErrorManager.recordOrThrowError(3133,
                 "SYSTEM ERROR: Sequence " + seqName.toUpperCase() + " not found");
         return;
      }
      if (seq.getMin() > newVal || newVal > seq.getMax())
      {
         ErrorManager.recordOrThrowError(3132,
                 "Cannot set sequence " + seq.getLegacyName().toUpperCase() + 
                 " beyond its max/min values");
         return;
      }
      
      /* H2 database server sets the current value of a sequence so as 
       * the following <code>next value</code> is called this newVal argument
       * is returned so the actual current value is newVal - increment. 
       * Progress 4GL instead sets the value and it is accessible via 
       * <code>(dynamic-)current-value</code> while 
       * <code>(dynamic-)next-value</code> will return newVal + increment.
       * The solution is to set the value already incremented into database. 
       * Note: newValFixed might be warped over long  limits! */
      long newValFixed = newVal + seq.getIncrement();

      Persistence persistence = ConnectionManager.getPersistence(ldbName);
      P2JDialect dialect = (P2JDialect) persistence.getDialect();
      String strQuery = dialect.getSequenceSetValString(seqName, newValFixed);

      try
      {
         persistence.executeSQL(strQuery);
      }
      catch (PersistenceException pe)
      {
         pe.printStackTrace();
      }
   }

   /**
    * Test whether the database contains a sequence. This is useful to
    * identify the correct database/sequences when a 4GL static function /
    * statement is called without the database parameter. 
    *
    * @param   seqName
    *          The name of the sequence.
    * @param   ldbName
    *          The database to which the sequence belongs to. If is null then
    *          the first connected database is assumed.
    *
    * @return  true if the sequence is defined in the database
    */
   @Override
   public boolean contains(String ldbName, String seqName)
   {
      // lazy initialization
      if (sequences == null)
      {
         sequences = new HashMap<String, LinkedList<String>>();
      }
         
      // make sure the structure is initialized for this database:
      LinkedList<String> crtSeqList = null;
      if (!sequences.containsKey(ldbName))
      {
         // create the list 
         crtSeqList = new LinkedList<String>();
         sequences.put(ldbName, crtSeqList);

         Persistence persistence = ConnectionManager.getPersistence(ldbName);
         try
         {
            ScrollableResults rs = persistence.executeSQLQuery(
                        "SELECT SEQUENCE_NAME " +
                        "FROM INFORMATION_SCHEMA.SEQUENCES;");
            while (rs.next())
            {
               crtSeqList.add(((String) rs.get(0)).toUpperCase());
            }
         }
         catch (PersistenceException pe)
         {
            pe.printStackTrace();
            return false;
         }
      }
      else
      {
         crtSeqList = sequences.get(ldbName);
      }

      // now check whether the sequence is in database
      return crtSeqList.contains(seqName.toUpperCase());
   }
}
