/*
** Module   : PostgreSQLSequenceHandler.java
** Abstract : Static class that handles sequence functions and statements 
**            from Progress / OpenEdge 10.x.
**
** Copyright (c) 2012-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 OM  20121116 Created initial version with CURRVAL, NEXTVAL and SETVAL support only.
** 002 SVL 20120331 Upgraded to Hibernate 4.
** 003 OM  20120712 Fixed extraction data from ResultSets. Added error handling
*/

package com.goldencode.p2j.persist.sequence;

import java.util.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.dialect.*;
import com.goldencode.p2j.util.*;
import org.hibernate.*;

/**
 * PostgreSQL implementation of sequences using a native SQL sequence
 * primitives queries.
 */
public class PostgreSQLSequenceHandler
extends SequenceHandler
{
   /**
    * The list of sequences grouped by databases.
    * It uses lazy initialization.
    */
   private HashMap<String, LinkedList<String>> sequences = null;

   /**
    * Queries the current value of a sequence.
    * @param   seqName
    *          The name of the sequence.
    * @param   ldbName
    *          The database to which the sequence belongs to. 
    *          It is assumed it is not null and valid.
    *
    * @return the current value of the sequence
    */
   @Override
   public int64 getCurrentValue(String seqName, String ldbName)
   {
      Sequence seq = SequenceManager.getSequence(ldbName, seqName);
      if (seq == null)
      {
         ErrorManager.recordOrThrowError(2913,
                 "Unknown sequence expression " + seqName.toUpperCase());
         return new int64();
      }
      
      // this will also create the instance if needed:
      Persistence persistence = ConnectionManager.getPersistence(ldbName);
      P2JDialect dialect = (P2JDialect) persistence.getDialect();
      String strQuery = dialect.getSequenceCurrValString(seqName);

      try
      {
         ScrollableResults rs = persistence.executeSQLQuery(strQuery);
         
         if (rs.next())
         {
            Object result = rs.get(0);
            if (result instanceof java.lang.Long)
            {
               // deprecated ?
               return new int64((java.lang.Long) rs.get(0));
            }
            else if (result instanceof java.math.BigInteger)
            {
               System.out.println(result);
               return new int64(((java.math.BigInteger) rs.get(0)).longValue());
            }
            else 
            {
               // don't know how to cast the result:
               System.err.println("** PostgreSQLSequenceHandler.getCurrentValue() cannot handle " +
                     result.getClass() + "result type.");
               return new int64();
            }
         }
      }
      catch (PersistenceException pe)
      {
         pe.printStackTrace();
      }

      // if here then some error occurred, return unknown value
      return new int64();
   }

   /**
    * Computes and returns the next value of a sequence.
    * @param   seqName
    *          The name of the sequence.
    * @param   ldbName
    *          The database to which the sequence belongs to. 
    *          It is assumed it is not null and valid.
    *
    * @return  the next value of the sequence or unknown value (?) if
    *          sequence/database not found or the sequence is not cycling
    *          and has passed the maximum/minimum value.
    */
   @Override
   public int64 getNextValue(String seqName, String ldbName)
   {
      Sequence seq = SequenceManager.getSequence(ldbName, seqName);
      if (seq == null)
      {
         ErrorManager.recordOrThrowError(2913,
                 "Unknown sequence expression " + seqName.toUpperCase());
         return new int64();
      }

      Persistence persistence = ConnectionManager.getPersistence(ldbName);
      P2JDialect dialect = (P2JDialect) persistence.getDialect();
      String strQuery = dialect.getSequenceNextValString(seqName);

      try
      {
         ScrollableResults rs = persistence.executeSQLQuery(strQuery);
         if (rs.next())
         {
            Object result = rs.get(0);
            if (result instanceof java.lang.Long)
            {
               // deprecated ?
               return new int64((java.lang.Long) rs.get(0));
            }
            else if (result instanceof java.math.BigInteger)
            {
               System.out.println(result);
               return new int64(((java.math.BigInteger) rs.get(0)).longValue());
            }
            else 
            {
               // don't know how to cast the result:
               System.err.println("** PostgreSQLSequenceHandler.getNextValue() cannot handle " +
                     result.getClass() + "result type.");
               return new int64();
            }
         }
      }
      catch (PersistenceException pe)
      {
         boolean outOfBoundsException = false;
         if (pe.getCause() != null)
         {
            String errMessage = pe.getCause().getMessage();
            if (errMessage.contains("reached maximum value") ||
                errMessage.contains("reached minimum value"))
            {
               outOfBoundsException = true;
            }
         }
         if (!outOfBoundsException)
         {
            // some other nasty thing has happen, report it:
            pe.printStackTrace();
         }
      }

      // if here then some error occurred, return unknown value
      return new int64();
   }

   /**
    * Initialize (reset) the current value of a sequence.
    *
    * @param   seqName
    *          The name of the sequence.
    * @param   ldbName
    *          The database to which the sequence belongs to. 
    *          It is assumed it is not null and valid.
    * @param   newVal
    *          The value the sequence will be reseted to.
    */
   @Override
   public void setValue(String seqName, String ldbName, long newVal)
   {
      Sequence seq = SequenceManager.getSequence(ldbName, seqName);
      if (seq == null)
      {
         ErrorManager.recordOrThrowError(3133,
                 "SYSTEM ERROR: Sequence " + seqName.toUpperCase() + " not found");
         return;
      }
      if (seq.getMin() > newVal || newVal > seq.getMax())
      {
         ErrorManager.recordOrThrowError(3132,
                 "Cannot set sequence " + seq.getLegacyName().toUpperCase() + 
                 " beyond its max/min values");
         return;
      }
      
      Persistence persistence = ConnectionManager.getPersistence(ldbName);
      P2JDialect dialect = (P2JDialect) persistence.getDialect();
      String strQuery = dialect.getSequenceSetValString(seqName, newVal);

      try
      {
         persistence.executeSQLQuery(strQuery);
      }
      catch (PersistenceException pe)
      {
         pe.printStackTrace();
      }
   }

   /**
    * Test whether the database contains a sequence. This is useful to
    * identify the correct database/sequences when a 4GL static function /
    * statement is called without the database parameter.
    *
    * @param   seqName
    *          The name of the sequence.
    * @param   ldbName
    *          The database to which the sequence belongs to. If is null then
    *          the first connected database is assumed.
    *
    * @return  true if the sequence is defined in the database
    */
   @Override
   public boolean contains(String ldbName, String seqName)
   {
      // lazy initialization
      if (sequences == null)
      {
         sequences = new HashMap<String, LinkedList<String>>();
      }
         
      // make sure the structure is initialized for this database:
      LinkedList<String> crtSeqList = null;
      if (!sequences.containsKey(ldbName))
      {
         // create the list 
         crtSeqList = new LinkedList<String>();
         sequences.put(ldbName, crtSeqList);

         Persistence persistence = ConnectionManager.getPersistence(ldbName);
         try
         {
            ScrollableResults rs = persistence.executeSQLQuery(
                        "SELECT relname " +
                        "FROM \"pg_class\" " +
                        "WHERE relkind='S';");
            while (rs.next())
            {
               crtSeqList.add((String) rs.get(0));
            }
         }
         catch (PersistenceException pe)
         {
            pe.printStackTrace();
            return false;
         }
      }
      else
      {
         crtSeqList = sequences.get(ldbName);
      }

      // now check whether the sequence is in database
      return crtSeqList.contains(seqName);
   }
}
